#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "bmp.h"

char *reverse(const char *text) {
    if (text == NULL)
        return NULL;
    size_t len = strlen(text);
    char *reversed = (char *)malloc(len + 1);
    if (reversed == NULL)
        return NULL;
    for (size_t i = 0; i < len; i++) {
        reversed[i] = toupper(text[len - i - 1]);
    }
    reversed[len] = '\0';
    return reversed;
}

char *vigenere_encrypt(const char *key, const char *text) {
    if (key == NULL || text == NULL) {
        return NULL;
    }

    for (size_t i = 0; key[i] != '\0'; i++) {
        if (!isalpha(key[i])) {
            return NULL;
        }
    }

    size_t key_len = strlen(key);
    size_t text_len = strlen(text);
    char *ciphertext = (char *)malloc(text_len + 1);
    if (ciphertext == NULL) {
        return NULL;
    }

    for (size_t i = 0, j = 0; i < text_len; i++) {
        char c = toupper(text[i]);
        if (c < 'A' || c > 'Z') {
            ciphertext[i] = c;
            continue;
        }
        char k = toupper(key[j % key_len]);
        ciphertext[i] = (c + k - 2 * 'A') % 26 + 'A';
        j++;
    }
    ciphertext[text_len] = '\0';
    return ciphertext;
}

char *vigenere_decrypt(const char *key, const char *text) {
    if (key == NULL || text == NULL) {
        return NULL;
    }

    for (size_t i = 0; key[i] != '\0'; i++) {
        if (!isalpha(key[i])) {
            return NULL;
        }
    }

    size_t key_len = strlen(key);
    size_t text_len = strlen(text);
    char *plaintext = (char *)malloc(text_len + 1);
    if (plaintext == NULL) {
        return NULL;
    }

    for (size_t i = 0, j = 0; i < text_len; i++) {
        char c = toupper(text[i]);
        if (c < 'A' || c > 'Z') {
            plaintext[i] = c;
            continue;
        }
        char k = toupper(key[j % key_len]);
        plaintext[i] = (c - k + 26) % 26 + 'A';
        j++;
    }
    plaintext[text_len] = '\0';
    return plaintext;
}

unsigned char *bit_encrypt(const char *text) {
    if (text == NULL) {
        return NULL;
    }

    size_t len = strlen(text);
    unsigned char *ciphertext = (unsigned char *)malloc(len + 1);
    if (ciphertext == NULL) {
        return NULL;
    }

    for (size_t i = 0; i < len; i++) {
        unsigned char c = text[i];
        unsigned char left = (c >> 4) & 0x0F;
        unsigned char right = c & 0x0F;   

        unsigned char new_left = ((left & 0x03) << 2) | ((left & 0x0C) >> 2);

        unsigned char result = (new_left ^ right);

        ciphertext[i] = result;
    }
    ciphertext[len] = '\0';
    return ciphertext;
}

char *bit_decrypt(const unsigned char *text) {
    if (text == NULL) {
        return NULL;
    }

    size_t len = strlen((const char *)text);
    char *plaintext = (char *)malloc(len + 1);
    if (plaintext == NULL) {
        return NULL;
    }

    for (size_t i = 0; i < len; i++) {
        unsigned char c = text[i];
        unsigned char left = (c >> 4) & 0x0F;
        unsigned char right = c & 0x0F;

        unsigned char new_left = ((left & 0x03) << 2) | ((left & 0x0C) >> 2);

        unsigned char original_right = (new_left ^ right);

        plaintext[i] = (new_left << 4) | original_right;
    }
    plaintext[len] = '\0';
    return plaintext;
}

unsigned char *bmp_encrypt(const char *key, const char *text) {
    if (key == NULL || text == NULL) {
        return NULL;
    }

    for (size_t i = 0; key[i] != '\0'; i++) {
        if (!isalpha(key[i])) {
            return NULL;
        }
    }

    char *reversed = reverse(text);
    if (reversed == NULL) {
        return NULL;
    }

    char *vigenere = vigenere_encrypt(key, reversed);
    free(reversed);
    if (vigenere == NULL) {
        return NULL;
    }

    unsigned char *ciphertext = bit_encrypt(vigenere);
    free(vigenere);
    return ciphertext;
}

char *bmp_decrypt(const char *key, const unsigned char *text) {
    if (key == NULL || text == NULL)
        return NULL;
    char *bit_decrypted = bit_decrypt(text);
    if (bit_decrypted == NULL)
        return NULL;

    char *vigenere_decrypted = vigenere_decrypt(key, bit_decrypted);

    free(bit_decrypted);
    if (vigenere_decrypted == NULL)
        return NULL;

    char *plaintext = reverse(vigenere_decrypted);
    free(vigenere_decrypted);
    return plaintext;
}